/*
  (c) 1998 netEstate (http://www.netestate.de/, info@netestate.de),
Michael Brunnbauer & Markus Hendel, Gesellschaft fuer Netzwerkkommunikation bR
Am Bluetenanger 8, 80995 Muenchen, Germany

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include "stringlib.h"

inline size_t string_add_buffer_size(size_t min_size)
{
    const size_t min_result = 128;
    
    // Um 50% groesser anlegen (schafft auch Platz fuer terminierende 0).
    min_size += min_size / 2;
    return min_size < min_result ? min_result : min_size;
}

inline string::string_holder *string::new_string_holder(size_t buffer_size)
{
    void *mem = operator new[](sizeof(string_holder) - 4 + buffer_size);
    string_mem += buffer_size;
    string_holder *holder = static_cast<string_holder *>(mem);
    holder->ref_count_ = 1;
    holder->buffer_size_ = buffer_size;
    return holder;
}

inline void string::construct(const char *begin, const char *end, size_t buffer_size)
{
    holder_ = new_string_holder(buffer_size);
    size_t str_size = end - begin;
    char *begin_ptr_ = holder_->buffer;
    char *end_ptr = begin_ptr_ + str_size;

    std::memcpy(begin_ptr_, begin, str_size);
    *end_ptr = '\0';

    holder_->end_ptr_ = end_ptr;

    begin_ = begin_ptr_;
    end_ = end_ptr;
}

inline void string::construct(const char *begin, const char *end)
{
    if (begin)
    {
        construct(begin, end, end - begin + 1);
    }
    else
    {
        begin_ = nullptr;
        end_ = nullptr;
        holder_ = nullptr;
    }
}

inline void string::construct(const char *s)
{
    if (s)
    {
        size_t size = std::strlen(s);
        construct(s, s + size, size + 1);
    }
    else
    {
        begin_ = nullptr;
        end_ = nullptr;
        holder_ = nullptr;
    }
}

inline string::string(const string &s, size_t size, size_t buffer_size)
{
    if (s.begin_)
    {
        construct(s.begin_, s.end_, buffer_size);
    }
    else
    {
        begin_ = nullptr;
        end_ = nullptr;
        holder_ = nullptr;
    }
}

string &string::operator=(const string &s)
{
    // Selbstzuweisung (this == &s) wird richtig behandelt, indem zuerst der
    // Referenzzhler von s inkrementiert wird.
    if (defined(s))
    {
        s.add_ref();
    }
    if (holder_)
    {
        release();
    }
    begin_ = s.begin_;
    end_ = s.end_;
    holder_ = s.holder_;
    return *this;
}

/* [] ueberladen */
char string::operator[](size_t pos) const
{
    if (!begin_)
    {
      log(LOG_FATAL,__FILE__,__LINE__,ERROR_UNSPECIFIED,
	  "called operator[] for string while string was NULL");
    }

    if (pos >= fast_size())
    {
      log(LOG_FATAL,__FILE__,__LINE__,ERROR_VALUE_OUT_OF_RANGE,
	  "Value was: %u / string: %s", pos, begin_);
    }
    return begin_[pos];
}

/* einzelne woerter */
string string::word(unsigned int i) const
{
    if (!begin_)
    {
        log(LOG_FATAL,__FILE__,__LINE__,ERROR_UNSPECIFIED,
          "called word() for string while string was NULL");
    }
    
    const char* rueck = begin_;
    const char* rueck_end = end_;
    
    while (i)
    {
        if (!(rueck=strchr(rueck,' ')))
        {
          log(LOG_FATAL,__FILE__,__LINE__,ERROR_VALUE_OUT_OF_RANGE,
              "Value was: %i /string: %s", i, begin_);
        }
        rueck++;
        i--;
    }
    
    const char *end = strchr(rueck,' ');
    if (end)
    {
        rueck_end = end;
    }
    
    return string(rueck, rueck_end);
}

/* wieviele woerter ? */
unsigned int string::words() const
{
  const char* zgr1;
  unsigned int i;

  if (!begin_)
  {
      log(LOG_FATAL,__FILE__,__LINE__,ERROR_UNSPECIFIED,
          "called words() for string while string was NULL");
  }
 
  zgr1=begin_; i=1;
  while ((zgr1=strchr(zgr1,' '))) { zgr1++; i++; };
  return i;
}

string operator+(const char *s1, const string &s2)
{
    if (!s1 || !defined(s2))
    {
        log(LOG_FATAL,__FILE__,__LINE__,ERROR_WRONG_ARGUMENT,
        "String argument for string operator+() was NULL");
    }
    size_t s1_size = std::strlen(s1);
    size_t s2_size = s2.end_ - s2.begin_;
    string s1_str(s1, s1_size, string_add_buffer_size(s1_size + s2_size));
    s1_str += s2;
    return s1_str;
}

string &string::operator+=(const string &s)
{
    if (!begin_)
    {
        return operator=(s);
    }
    
    if (!defined(s))
    {
        log(LOG_FATAL,__FILE__,__LINE__,ERROR_WRONG_ARGUMENT,
            "String argument for string operator+=() was NULL\n");
    }

    size_t this_size = fast_size();
    size_t s_size = s.fast_size();
    if (end_ != holder_->end_ptr_ || this_size + s_size + 1 > holder_->buffer_size_)
    {
        // Dies ist ein substring oder Puffer zu klein, wir mssen kopieren.
        string_holder *new_holder = new_string_holder(string_add_buffer_size(this_size + s_size));

        char *new_begin = new_holder->buffer;
        std::memcpy(new_begin, begin_, this_size);

        // Alten holder freigeben.
        release();
        
        // Mit neuem holder weitermachen.
        begin_ = new_begin;
        end_ = new_begin + this_size;
        holder_ = new_holder;
    }
    // Ok, jetzt ist genug Platz. s dranhaengen
    std::memcpy(end_, s.begin_, s_size);
    end_[s_size] = '\0';
    end_ += s_size;
    holder_->end_ptr_ = end_;
    return *this;
}

void string::append(const char *s, size_t size)
{
    if (!begin_ || !s)
    {
        log(LOG_FATAL,__FILE__,__LINE__,ERROR_WRONG_ARGUMENT,
            "String argument for string append() was NULL");
    }

    size_t this_size = fast_size();
    if (end_ != holder_->end_ptr_ || this_size + size + 1 > holder_->buffer_size_)
    {
        // Dies ist ein substring oder Puffer zu klein, wir mssen kopieren.
        string_holder *new_holder = new_string_holder(string_add_buffer_size(this_size + size));

        char *new_begin = new_holder->buffer;
        std::memcpy(new_begin, begin_, this_size);

        // Alten holder freigeben.
        release();
        
        // Mit neuem holder weitermachen.
        begin_ = new_begin;
        end_ = new_begin + this_size;
        holder_ = new_holder;
    }
    // Ok, jetzt ist genug Platz. s dranhaengen
    std::memcpy(end_, s, size);
    end_[size] = '\0';
    end_ += size;
    holder_->end_ptr_ = end_;
}

/* == ueberladen */
bool operator==(const string &s1, const string &s2)
{
    if (!defined(s1) || !defined(s2))
    {
        log(LOG_FATAL,__FILE__,__LINE__,ERROR_WRONG_ARGUMENT,
            "String argument for string operator==() was NULL");
    }

    size_t s1_size = s1.fast_size();
    size_t s2_size = s2.fast_size();

    if (s1_size != s2_size)
    {
        return false;
    }

    return std::strncmp(s1.begin(), s2.begin(), s1_size) == 0;
}

bool operator==(const string &s1, const char *s2)
{
    if (!defined(s1) || !s2)
    {
        log(LOG_FATAL,__FILE__,__LINE__,ERROR_WRONG_ARGUMENT,
            "String argument for string operator==() was NULL");
    }

    size_t s1_size = s1.size();

    // FIXME: nicht sehr schnell: s2 wird ggf. doppelt gescannt, es gibt
    //        aber keine std-funktion die das schneller koennte.
    if (std::strncmp(s1.begin(), s2, s1_size) != 0)
    {
        return false;
    }
    return s1_size == std::strlen(s2);
}
