/*
  (c) 1998 netEstate (http://www.netestate.de/, info@netestate.de),
Michael Brunnbauer & Markus Hendel, Gesellschaft fuer Netzwerkkommunikation bR
Am Bluetenanger 8, 80995 Muenchen, Germany

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#ifndef LIBS_STRINGLIB_H_INCLUDED
#define LIBS_STRINGLIB_H_INCLUDED

#ifndef PRIVATE_INCLUDE_general
#include "general.h"
#define PRIVATE_INCLUDE_general
#endif
#ifndef PRIVATE_INCLUDE_errorlib
#include "errorlib.h"
#define PRIVATE_INCLUDE_errorlib
#endif

#ifndef NOSERIALIZE
#ifndef PRIVATE_INCLUDE_serialize
#include "serialize.h"
#define PRIVATE_INCLUDE_serialize
#endif
#endif

/* FIXME: gcc version 2.95.2 broken
#ifndef INCLUDE_iostream
#include <iostream.h>
#define INCLUDE_iostream
#endif
#ifndef INCLUDE_iomanip
#include <iomanip.h>
#define INCLUDE_iomanip
#endif
#ifndef INCLUDE_fstream
#include <fstream.h>
#define INCLUDE_fstream
#endif
*/

// Bug in g++: Kann nullptr nicht korrekt definieren.
#ifndef nullptr
#define nullptr 0
#endif

#include <cstring>

extern unsigned long string_mem;

class string
{
public:

    /* konstruktor */
    string();
    
    /* destruktor */
    ~string();
    
    /* copy-konstruktor */
    string(const string &s);
    
    /* konversion von const char* nach string */
    string(const char *);
    
    /* iterator range konstruktor */
    string(const char *begin, const char *end);

    /* = ueberladen */
    string &operator=(const string &s);
    
    /* = ueberladen */
    // string &operator=(const char *s);
    // nicht benoetigt, muss hier sowieso eine Kopie machen.
    
    /* += ueberladen */
    string &operator+=(const string &s);
    string &operator+=(const char *s);
    void append(const char *s, size_t size);
    
    /* [] ueberladen, gibt _keine_ Referenz auf das char zurueck */
    char operator[](size_t pos) const;
    
    /* konversion von string nach null-terminiertem const char* */
    const char *c_str();

    /* lock() gibt einen beschreibbaren Zeiger auf das erste Zeichen den string zurueck.
       Achtung: Es darf keine andere Funktion auer c_str() aufgerufen werden, bevor wieder unlock()
       aufgerufen wurde. unlock() muss nur aufgerufen werden, falls terminierende '\0'-zeichen
       in den string geschrieben wurden. */
    char *lock();
    void unlock();

    /* begin-iterator */
    const char *begin() const;

    /* end-iterator */
    const char *end() const;

    /* laenge */
    size_t size() const;

    /* laenge, falls defined */
    size_t fast_size() const;

    /* substrings */
    string substring(size_t start_pos, size_t end_pos) const;
    
    /* sonstiges */
    string word(unsigned int) const;
    unsigned int words() const;

    void swap(string &s);

    friend string operator+(const char *s1, const string &s2);

private:

    string(const string &s, size_t size, size_t buffer_size);

    struct string_holder
    {
        size_t  ref_count_;
        size_t  buffer_size_; // gesamte Puffergroesse.
        char   *end_ptr_;     // zeigt auf das erste '\0' im Puffer
        char    buffer[4];    // Puffer, 4 ist wahrscheinliche minimale ausrichtung.
    };

    static string_holder *new_string_holder(size_t buffer_size);


    void add_ref() const; // logisch const
    void release() const; // logisch const
    void dup();

    void construct(const char *begin, const char *end, size_t buffer_size);
    void construct(const char *begin, const char *end);
    void construct(const char *s);

    char          *begin_; // Zeiger auf Anfang des strings im Puffer.
    char          *end_;   // Zeiger auf Anfang des strings im Puffer.
    string_holder *holder_;
};

template<typename T>
inline void swap_basic(T &t1, T &t2)
{
    T tmp = t1;
    t1 = t2;
    t2 = tmp;
}

/* ausgabe in streams */
/* ostream& operator<<(ostream&, string&); */

/* +, ==, !=, += ueberladen */

inline void string::swap(string &s)
{
    swap_basic(begin_, s.begin_);
    swap_basic(end_, s.end_);
    swap_basic(holder_, s.holder_);
}

inline bool defined(const string &s)
{
    return s.begin();
}

inline void string::add_ref() const
{
    ++ holder_->ref_count_;
}

inline void string::release() const
{
    if (!--holder_->ref_count_)
    {
        string_mem -= holder_->buffer_size_;
        operator delete[](holder_);
    }
}

inline const char *string::begin() const
{
    return begin_;
}

inline const char *string::end() const
{
    return end_;
}

inline size_t string::fast_size() const
{
    return end_ - begin_;
}

inline string::string(const char *s)
{
    construct(s);
}

inline string::string(const char *begin, const char *end)
{
    construct(begin, end);
}

inline void string::dup()
{
    operator=(string(begin_, end_));
}

inline const char *string::c_str()
{
    if (!begin_ || end_ == holder_->end_ptr_)
    {
        return begin_;
    }
    dup();
    return begin_; 
}

inline char *string::lock()
{
    if (!begin_ || (end_ == holder_->end_ptr_ && holder_->ref_count_ == 1))
    {
        return begin_;
    }
    dup();
    return begin_;
}

inline void string::unlock()
{
    // Laenge anpassen, falls '\0'-chars in den string geschrieben wurden
    // (z.B. durch strtok()).
    if (begin_)
    {
        end_ = begin_ + std::strlen(begin_);
        holder_->end_ptr_ = end_;
    }
}

inline string::string()
:   begin_(nullptr),
    end_(nullptr),
    holder_(nullptr)
{ }

/* destruktor */
inline string::~string()
{
    if (holder_)
    {
        release();
    }
}

/* copy-konstruktor */
inline string::string(const string &s)
{
    begin_ = s.begin_;
    end_ = s.end_;
    holder_ = s.holder_;
    if (holder_)
    {
        add_ref();
    }
}

inline size_t string::size() const
{
    if (!begin_)
    {
        return 0;
    }
    return end_ - begin_;
}

inline string operator+(string s1, const string &s2)
{
    s1 += s2;
    return s1;
}

string operator+(const char *s1, const string &s2);

inline string operator+(string s1, const char *s2)
{
    s1 += s2;
    return s1;
}

bool operator==(const string &s1, const string &s2);

inline bool operator!=(const string &s1, const string &s2)
{
    return !(s1 == s2);
}

bool operator==(const string &s1, const char *s2);

inline bool operator!=(const string &s1, const char *s2)
{
    return !(s1 == s2);
}

inline bool operator==(const char *s1, const string &s2)
{
    return s2 == s1;
}

inline bool operator!=(const char *s1, const string &s2)
{
    return !(s1 == s2);
}

inline string &string::operator+=(const char *s)
{
    // FIXME: Optimieren: Kein temp. Objekt erstellen.
    return operator+=(string(s));
}

#endif // ndef LIBS_STRINGLIB_H_INCLUDED
